@@ -1,11 +1,8 @@
 # frozen_string_literal: true
 require 'rbconfig'
 require 'parallel/version'
-require 'parallel/processor_count'
 
 module Parallel
-  extend ProcessorCount
-
   Stop = Object.new.freeze
 
   class DeadWorker < StandardError
@@ -27,7 +24,7 @@ module Parallel
     attr_reader :backtrace
 
     def initialize(original)
-      super "#{original.class}: #{original.message}"
+      super("#{original.class}: #{original.message}")
       @backtrace = original.backtrace
     end
   end
@@ -303,8 +300,47 @@ module Parallel
       map(array, options.merge(with_index: true), &block)
     end
 
-    def flat_map(*args, &block)
-      map(*args, &block).flatten(1)
+    def flat_map(...)
+      map(...).flatten(1)
+    end
+
+    def filter_map(...)
+      map(...).compact
+    end
+
+    # Number of physical processor cores on the current system.
+    def physical_processor_count
+      @physical_processor_count ||= begin
+        ppc =
+          case RbConfig::CONFIG["target_os"]
+          when /darwin[12]/
+            IO.popen("/usr/sbin/sysctl -n hw.physicalcpu").read.to_i
+          when /linux/
+            cores = {} # unique physical ID / core ID combinations
+            phy = 0
+            File.read("/proc/cpuinfo").scan(/^physical id.*|^core id.*/) do |ln|
+              if ln.start_with?("physical")
+                phy = ln[/\d+/]
+              elsif ln.start_with?("core")
+                cid = "#{phy}:#{ln[/\d+/]}"
+                cores[cid] = true unless cores[cid]
+              end
+            end
+            cores.count
+          when /mswin|mingw/
+            physical_processor_count_windows
+          else
+            processor_count
+          end
+        # fall back to logical count if physical info is invalid
+        ppc > 0 ? ppc : processor_count
+      end
+    end
+
+    # Number of processors seen by the OS or value considering CPU quota if the process is inside a cgroup,
+    # used for process scheduling
+    def processor_count
+      @processor_count ||= Integer(ENV['PARALLEL_PROCESSOR_COUNT'] || available_processor_count)
     end
 
     def worker_number
@@ -318,8 +354,35 @@ module Parallel
 
     private
 
+    def physical_processor_count_windows
+      # Get-CimInstance introduced in PowerShell 3 or earlier: https://learn.microsoft.com/en-us/previous-versions/powershell/module/cimcmdlets/get-ciminstance?view=powershell-3.0
+      result = run(
+        'powershell -command "Get-CimInstance -ClassName Win32_Processor -Property NumberOfCores ' \
+        '| Select-Object -Property NumberOfCores"'
+      )
+      if !result || $?.exitstatus != 0
+        # fallback to deprecated wmic for older systems
+        result = run("wmic cpu get NumberOfCores")
+      end
+      if !result || $?.exitstatus != 0
+        # Bail out if both commands returned something unexpected
+        warn "guessing pyhsical processor count"
+        processor_count
+      else
+        # powershell: "\nNumberOfCores\n-------------\n            4\n\n\n"
+        # wmic:       "NumberOfCores  \n\n4              \n\n\n\n"
+        result.scan(/\d+/).map(&:to_i).reduce(:+)
+      end
+    end
+
+    def run(command)
+      IO.popen(command, &:read)
+    rescue Errno::ENOENT
+      # Ignore
+    end
+
     def add_progress_bar!(job_factory, options)
-      if progress_options = options[:progress]
+      if (progress_options = options[:progress])
         raise "Progressbar can only be used with array like items" if job_factory.size == Float::INFINITY
         require 'ruby-progressbar'
 
@@ -348,7 +411,7 @@ module Parallel
       results = []
       exception = nil
       begin
-        while set = job_factory.next
+        while (set = job_factory.next)
           item, index = set
           results << with_instrumentation(item, index, options) do
             call_with_index(item, index, options, &block)
@@ -371,7 +434,7 @@ module Parallel
       in_threads(options) do |worker_num|
         self.worker_number = worker_num
         # as long as there are more jobs, work on one of them
-        while !exception && set = job_factory.next
+        while !exception && (set = job_factory.next)
           begin
             item, index = set
             result = with_instrumentation item, index, options do
@@ -415,7 +478,7 @@ module Parallel
 
       # start
       ractors.dup.each do |ractor|
-        if set = job_factory.next
+        if (set = job_factory.next)
           item, index = set
           instrument_start item, index, options
           ractor.send [callback, item, index]
@@ -426,7 +489,7 @@ module Parallel
       end
 
       # replace with new items
-      while set = job_factory.next
+      while (set = job_factory.next)
         item_next, index_next = set
         done, (exception, result, item, index) = Ractor.select(*ractors)
         if exception
@@ -600,13 +663,46 @@ module Parallel
     end
 
     def instrument_finish(item, index, result, options)
-      return unless on_finish = options[:finish]
+      return unless (on_finish = options[:finish])
+      return instrument_finish_in_order(item, index, result, options) if options[:finish_in_order]
       options[:mutex].synchronize { on_finish.call(item, index, result) }
     end
 
+    # yield results in the order of the input items
+    # needs to use `options` to store state between executions
+    # needs to use `done` index since a nil result would also be valid
+    def instrument_finish_in_order(item, index, result, options)
+      options[:mutex].synchronize do
+        # initialize our state
+        options[:finish_done] ||= []
+        options[:finish_expecting] ||= 0 # we wait for item at index 0
+
+        # store current result
+        options[:finish_done][index] = [item, result]
+
+        # yield all results that are now in order
+        break unless index == options[:finish_expecting]
+        index.upto(options[:finish_done].size).each do |i|
+          break unless (done = options[:finish_done][i])
+          options[:finish_done][i] = nil # allow GC to free this item and result
+          options[:finish].call(done[0], i, done[1])
+          options[:finish_expecting] += 1
+        end
+      end
+    end
+
     def instrument_start(item, index, options)
-      return unless on_start = options[:start]
+      return unless (on_start = options[:start])
       options[:mutex].synchronize { on_start.call(item, index) }
     end
+
+    def available_processor_count
+      gem 'concurrent-ruby', '>= 1.3.4'
+      require 'concurrent-ruby'
+      Concurrent.available_processor_count.floor
+    rescue LoadError
+      require 'etc'
+      Etc.nprocessors
+    end
   end
 end
